# mypy: ignore-errors
# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off
# isort: skip_file
from .....schema_classes import DynamicFormAssignmentClass
from .....schema_classes import FormActorAssignmentClass
from .....schema_classes import FormInfoClass
from .....schema_classes import FormPromptClass
from .....schema_classes import FormPromptTypeClass
from .....schema_classes import FormTypeClass
from .....schema_classes import StructuredPropertyParamsClass


DynamicFormAssignment = DynamicFormAssignmentClass
FormActorAssignment = FormActorAssignmentClass
FormInfo = FormInfoClass
FormPrompt = FormPromptClass
FormPromptType = FormPromptTypeClass
FormType = FormTypeClass
StructuredPropertyParams = StructuredPropertyParamsClass

# fmt: on
